@@ -16,20 +16,25 @@ from crytic_compile import CryticCompile, InvalidCompilation
 from .utils import color_good, color_warn
 from .sevm import *
 
-sys.set_int_max_str_digits(0)
+if hasattr(sys, 'set_int_max_str_digits'): # Python verion >=3.8.14, >=3.9.14, >=3.10.7, or >=3.11
+    sys.set_int_max_str_digits(0)
 
-def parse_args() -> argparse.Namespace:
+def parse_args(args) -> argparse.Namespace:
     parser = argparse.ArgumentParser(prog='halmos', epilog='For more information, see https://github.com/a16z/halmos')
 
     parser.add_argument('target', metavar='TARGET_DIRECTORY', nargs='?', default=os.getcwd(), help='source root directory (default: current directory)')
     parser.add_argument('--contract', metavar='CONTRACT_NAME', help='run tests in the given contract only')
     parser.add_argument('--function', metavar='FUNCTION_NAME_PREFIX', default='test', help='run tests matching the given prefix only (default: %(default)s)')
 
+    parser.add_argument('--bytecode', metavar='HEX_STRING', help='execute the given bytecode')
+
     parser.add_argument('--loop', metavar='MAX_BOUND', type=int, default=2, help='set loop unrolling bounds (default: %(default)s)')
     parser.add_argument('--width', metavar='MAX_WIDTH', type=int, help='set the max number of paths')
     parser.add_argument('--depth', metavar='MAX_DEPTH', type=int, help='set the max path length')
     parser.add_argument('--array-lengths', metavar='NAME1=LENGTH1,NAME2=LENGTH2,...', help='set the length of dynamic-sized arrays including bytes and string (default: loop unrolling bound)')
 
+    parser.add_argument('--symbolic-jump', action='store_true', help='support symbolic jump destination (experimental)')
+
     parser.add_argument('--no-smt-add',          action='store_true', help='do not interpret `+`')
     parser.add_argument('--no-smt-sub',          action='store_true', help='do not interpret `-`')
     parser.add_argument('--no-smt-mul',          action='store_true', help='do not interpret `*`')
@@ -49,92 +54,202 @@ def parse_args() -> argparse.Namespace:
 
     cryticparser.init(parser)
 
-    return parser.parse_args()
+    return parser.parse_args(args)
 
-def add_srcmap(ops: List[Opcode], srcmap: List[str], srcs: Dict) -> None:
-    fpath = {} # file id -> file path
-    for src in srcs:
-        fpath[srcs[src]['id']] = src
-    #   print(src, srcs[src]['id'])
+def str_abi(item: Dict) -> str:
+    def str_tuple(args: List) -> str:
+        ret = []
+        for arg in args:
+            typ = arg['type']
+            if typ == 'tuple':
+            #   ret.append(str_tuple(arg['components']))
+                ret.append(typ) # crytic-compile bug
+            else:
+                ret.append(typ)
+        return '(' + ','.join(ret) + ')'
+    if item['type'] != 'function': raise ValueError(item)
+    return item['name'] + str_tuple(item['inputs'])
 
-    start, length, fileid, jump, mdepth = 0, 0, 0, '-', 0
-    for idx, sm in enumerate(srcmap):
-        arr = sm.split(':') + ['']*5
-        start  = int(arr[0]) if arr[0] != '' else start
-        length = int(arr[1]) if arr[1] != '' else length
-        srcidx = int(arr[2]) if arr[2] != '' else srcidx
-        jump   =     arr[3]  if arr[3] != '' else jump
-        mdepth = int(arr[4]) if arr[4] != '' else mdepth
+def find_abi(abi: List, funname: str, funsig: str) -> Dict:
+    for item in abi:
+        if item['type'] == 'function' and item['name'] == funname and str_abi(item) == funsig:
+            return item
+    raise ValueError('Not found', abi, funsig)
 
-        if srcidx in fpath:
-            with open(f'{args.target}/{fpath[srcidx]}') as f:
-                f.seek(start)
-                srctext = repr(f.read(length))
+def mk_calldata(abi: List, funname: str, funsig: str, arrlen: Dict, args: argparse.Namespace, cd: List, dyn_param_size: List[str]) -> None:
+    item = find_abi(abi, funname, funsig)
+    tba = []
+    offset = 0
+    for param in item['inputs']:
+        param_name = param['name']
+        param_type = param['type']
+        if param_type == 'tuple':
+            raise ValueError('Not supported', param_type) # TODO: support struct types
+        elif param_type == 'bytes' or param_type == 'string':
+            tba.append((4+offset, param)) # wstore(cd, 4+offset, 32, BitVecVal(<?offset?>, 256))
+            offset += 32
+        elif param_type.endswith('[]'):
+            raise ValueError('Not supported variable sized arrays', param_type)
         else:
-            srctext = '<generated>'
+            match = re.search(r'(u?int[0-9]*|address|bool|bytes[0-9]+)(\[([0-9]+)\])?', param_type)
+            if not match: raise ValueError('Unknown type', param_type)
+            typ = match.group(1)
+            dim = match.group(3)
+            if dim: # array
+                for idx in range(int(dim)):
+                    wstore(cd, 4+offset, 32, BitVec(f'p_{param_name}[{idx}]_{typ}', 256))
+                    offset += 32
+            else: # primitive
+                wstore(cd, 4+offset, 32, BitVec(f'p_{param_name}_{typ}', 256))
+                offset += 32
 
-        ops[idx].sm = SrcMap(srctext, jump, mdepth)
+    for loc_param in tba:
+        loc   = loc_param[0]
+        param = loc_param[1]
+        param_name = param['name']
+        param_type = param['type']
+
+        if param_name not in arrlen:
+            size = args.loop
+            if args.debug: print(f'warn: size of {param_name} not given, using default value {size}')
+        else:
+            size = arrlen[param_name]
+
+        dyn_param_size.append(f'|{param_name}|={size}')
+
+        if param_type == 'bytes' or param_type == 'string':
+            # head
+            wstore(cd, loc, 32, BitVecVal(offset, 256))
+            # tail
+            size_pad_right = int((size + 31) / 32) * 32
+            wstore(cd, 4+offset, 32, BitVecVal(size, 256))
+            offset += 32
+            if size_pad_right > 0:
+                wstore(cd, 4+offset, size_pad_right, BitVec(f'p_{param_name}_{param_type}', 8*size_pad_right))
+                offset += size_pad_right
+        else:
+            raise ValueError('not feasible')
+
+def is_stop_or_return(opcode: Byte) -> bool:
+    return is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN]
+
+def decode_hex(hexcode: str) -> Tuple[List[Opcode], List[Any]]:
+    if hexcode.startswith('0x'):
+        hexcode = hexcode[2:]
+    if len(hexcode) % 2 != 0: raise ValueError(hexcode)
+    (ops, code) = decode(BitVecVal(int(hexcode, 16), (len(hexcode) // 2) * 8))
+    pgm = ops_to_pgm(ops)
+    return (pgm, code)
+
+def mk_callvalue() -> Word:
+    return BitVec('msg_value', 256)
+
+def mk_balance() -> Word:
+    return BitVec('this_balance', 256)
+
+def mk_caller(solver) -> Word:
+    caller = BitVec('msg_sender', 256)
+    solver.add(Extract(255, 160, caller) == BitVecVal(0, 96))
+    return caller
+
+def mk_this(solver) -> Word:
+    this = BitVec('this_address', 256)
+    solver.add(Extract(255, 160, this) == BitVecVal(0, 96))
+    return this
+
+def mk_solver(args: argparse.Namespace):
+    solver = SolverFor('QF_AUFBV') # quantifier-free bitvector + array theory; https://smtlib.cs.uiowa.edu/logics.shtml
+    solver.set(timeout=args.solver_timeout_branching)
+    return solver
+
+def run_bytecode(hexcode: str, args: argparse.Namespace, options: Dict) -> List[Exec]:
+    (pgm, code) = decode_hex(hexcode)
+
+    storage = {}
+
+    solver = mk_solver(args)
+
+    balance = mk_balance()
+    callvalue = mk_callvalue()
+    caller = mk_caller(solver)
+    this = mk_this(solver)
+
+    sevm = SEVM(options)
+    ex = sevm.mk_exec(
+        pgm       = { this: pgm },
+        code      = { this: code },
+        storage   = { this: storage },
+        balance   = { this: balance },
+        calldata  = [],
+        callvalue = callvalue,
+        caller    = caller,
+        this      = this,
+        symbolic  = True,
+        solver    = solver,
+    )
+    (exs, _) = sevm.run(ex)
+
+    models = []
+    for idx, ex in enumerate(exs):
+        opcode = ex.pgm[ex.this][ex.pc].op[0]
+        if is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN, EVM.REVERT]:
+            gen_model(args, models, idx, ex)
+            print(f'Final opcode: {opcode.as_long()} | Return data: {ex.output} | Input example: {models[-1][0]}')
+        else:
+            print(color_warn('Not supported: ' + opcode + ' ' + ex.error))
+        if args.verbose >= 1:
+            print(f'# {idx+1} / {len(exs)}')
+            print(ex)
+
+    return exs
 
 def setup(
     hexcode: str,
-    abi: Dict,
-    srcmap: List[str],
-    srcs: Dict,
-    args: argparse.Namespace,
+    abi: List,
+    setup_name: str,
     setup_sig: str,
+    setup_selector: str,
+    arrlen: Dict,
+    args: argparse.Namespace,
     options: Dict
 ) -> Exec:
-    # bytecode
-    (ops, code) = decode(hexcode)
-    pgm = ops_to_pgm(ops)
-    add_srcmap(ops, srcmap, srcs)
+    (pgm, code) = decode_hex(hexcode)
 
-    # solver
-    solver = SolverFor('QF_AUFBV') # quantifier-free bitvector + array theory; https://smtlib.cs.uiowa.edu/logics.shtml
-    solver.set(timeout=args.solver_timeout_branching)
+    solver = mk_solver(args)
 
-    # storage
-    storage = {}
-
-    # caller
-    caller = BitVec('msg_sender', 256)
-    solver.add(Extract(255, 160, caller) == BitVecVal(0, 96))
-
-    # this
-    this = BitVec('this_address', 256)
-    solver.add(Extract(255, 160, this) == BitVecVal(0, 96))
-
-    # run setup if any
+    this = mk_this(solver)
 
     sevm = SEVM(options)
 
     setup_ex = sevm.mk_exec(
         pgm       = { this: pgm },
         code      = { this: code },
-        storage   = { this: storage },
+        storage   = { this: {} },
         balance   = { this: con(0) },
-        calldata  = [None] * 4,
+        calldata  = [],
         callvalue = con(0),
-        caller    = caller,
+        caller    = mk_caller(solver),
         this      = this,
         symbolic  = False,
         solver    = solver,
     )
 
     if setup_sig:
-        wstore(setup_ex.calldata, 0, 4, BitVecVal(setup_sig, 32))
+        wstore(setup_ex.calldata, 0, 4, BitVecVal(setup_selector, 32))
+        dyn_param_size = [] # TODO: propagate to run
+        mk_calldata(abi, setup_name, setup_sig, arrlen, args, setup_ex.calldata, dyn_param_size)
 
         (setup_exs, setup_steps) = sevm.run(setup_ex)
 
-        if len(setup_exs) != 1:
+        setup_exs = list(filter(lambda ex: is_stop_or_return(ex.pgm[ex.this][ex.pc].op[0]) and not ex.failed, setup_exs))
+
+        if len(setup_exs) == 0: raise ValueError('setUp() failed')
+        if len(setup_exs) > 1:
             if args.debug: print('\n'.join(map(str, setup_exs)))
             raise ValueError('multiple paths exist in setUp()')
 
         setup_ex = setup_exs[0]
 
-        setup_opcode = setup_ex.pgm[setup_ex.this][setup_ex.pc].op[0]
-        if (setup_opcode != 'STOP' and setup_opcode != 'RETURN') or setup_ex.failed: raise ValueError('setUp() failed')
-
         if args.verbose >= 2:
             print(setup_ex)
 
@@ -142,7 +257,7 @@ def setup(
 
 def run(
     setup_ex: Exec,
-    abi: Dict,
+    abi: List,
     funname: str,
     funsig: str,
     funselector: str,
@@ -154,81 +269,24 @@ def run(
     # calldata
     #
 
-    f_cd = Function('cd', BitVecSort(256), BitVecSort(8))
-    cdsize = 10000
     cd = []
-    for i in range(cdsize):
-        cd.append(f_cd(con(i)))
 
     wstore(cd, 0, 4, BitVecVal(funselector, 32))
 
     dyn_param_size = []
-
-    for item in abi:
-        if item['type'] == 'function' and item['name'] == funname:
-            tba = []
-            offset = 0
-            for param in item['inputs']:
-                param_name = param['name']
-                param_type = param['type']
-                if param_type == 'tuple':
-                    raise ValueError('Not supported', param_type) # TODO: support struct types
-                elif param_type == 'bytes' or param_type == 'string':
-                    tba.append((4+offset, param)) # wstore(cd, 4+offset, 32, BitVecVal(<?offset?>, 256))
-                    offset += 32
-                elif param_type.endswith('[]'):
-                    raise ValueError('Not supported variable sized arrays', param_type)
-                else:
-                    match = re.search(r'(u?int[0-9]*|address|bool|bytes[0-9]+)(\[([0-9]+)\])?', param_type)
-                    if not match: raise ValueError('Unknown type', param_type)
-                    typ = match.group(1)
-                    dim = match.group(3)
-                    if dim: # array
-                        for idx in range(int(dim)):
-                            wstore(cd, 4+offset, 32, BitVec(f'p_{param_name}[{idx}]_{typ}', 256))
-                            offset += 32
-                    else: # primitive
-                        wstore(cd, 4+offset, 32, BitVec(f'p_{param_name}_{typ}', 256))
-                        offset += 32
-
-            for loc_param in tba:
-                loc   = loc_param[0]
-                param = loc_param[1]
-                param_name = param['name']
-                param_type = param['type']
-
-                if param_name not in arrlen:
-                    size = args.loop
-                    if args.debug: print(f'warn: size of {param_name} not given, using default value {size}')
-                else:
-                    size = arrlen[param_name]
-
-                dyn_param_size.append(f'|{param_name}|={size}')
-
-                if param_type == 'bytes' or param_type == 'string':
-                    # head
-                    wstore(cd, loc, 32, BitVecVal(offset, 256))
-                    # tail
-                    size_pad_right = int((size + 31) / 32) * 32
-                    wstore(cd, 4+offset, 32, BitVecVal(size, 256))
-                    offset += 32
-                    if size_pad_right > 0:
-                        wstore(cd, 4+offset, size_pad_right, BitVec(f'p_{param_name}_{param_type}', 8*size_pad_right))
-                        offset += size_pad_right
-                else:
-                    raise ValueError('not feasible')
+    mk_calldata(abi, funname, funsig, arrlen, args, cd, dyn_param_size)
 
     #
     # callvalue
     #
 
-    callvalue = BitVec('msg_value', 256)
+    callvalue = mk_callvalue()
 
     #
     # balance
     #
 
-    balance = BitVec('this_balance', 256)
+    balance = mk_balance()
 
     #
     # run
@@ -246,7 +304,7 @@ def run(
         pgm       = setup_ex.pgm.copy(), # shallow copy
         code      = setup_ex.code.copy(), # shallow copy
         storage   = deepcopy(setup_ex.storage),
-        balance   = { setup_ex.this: sevm.arith('ADD', balance, callvalue) },
+        balance   = { setup_ex.this: sevm.arith(EVM.ADD, balance, callvalue) },
         #
         calldata  = cd,
         callvalue = callvalue,
@@ -277,12 +335,12 @@ def run(
     stuck = []
     for idx, ex in enumerate(exs):
         opcode = ex.pgm[ex.this][ex.pc].op[0]
-        if opcode == 'STOP' or opcode == 'RETURN':
+        if is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN]:
             if ex.failed:
                 gen_model(args, models, idx, ex)
             else:
                 normal += 1
-        elif opcode == 'REVERT':
+        elif is_bv_value(opcode) and opcode.as_long() == EVM.REVERT:
             # Panic(1) # bytes4(keccak256("Panic(uint256)")) + bytes32(1)
             if ex.output == int('4e487b71' + '0000000000000000000000000000000000000000000000000000000000000001', 16): # 152078208365357342262005707660225848957176981554335715805457651098985835139029979365377
                 gen_model(args, models, idx, ex)
@@ -316,7 +374,7 @@ def run(
     # print post-states
     if args.verbose >= 2:
         for idx, ex in enumerate(exs):
-            if args.print_revert or (ex.pgm[ex.this][ex.pc].op[0] != 'REVERT' and not ex.failed):
+            if args.print_revert or (is_stop_or_return(ex.pgm[ex.this][ex.pc].op[0]) and not ex.failed):
                 print(f'# {idx+1} / {len(exs)}')
                 print(ex)
 
@@ -411,6 +469,23 @@ def str_model(model, args: argparse.Namespace) -> str:
     else:
         return '[' + ', '.join(sorted(map(lambda decl: f'{decl} = {model[decl]}', filter(select, model)))) + ']'
 
+def mk_options(args: argparse.Namespace) -> Dict:
+    return {
+        'target': args.target,
+        'verbose': args.verbose,
+        'debug': args.debug,
+        'log': args.log,
+        'add': not args.no_smt_add,
+        'sub': not args.no_smt_sub,
+        'mul': not args.no_smt_mul,
+        'div': args.smt_div,
+        'divByConst': args.smt_div_by_const,
+        'modByConst': args.smt_mod_by_const,
+        'expByConst': args.smt_exp_by_const,
+        'timeout': args.solver_timeout_branching,
+        'sym_jump': args.symbolic_jump,
+    }
+
 def main() -> int:
     #
     # z3 global options
@@ -424,21 +499,9 @@ def main() -> int:
     # command line arguments
     #
 
-    args = parse_args()
+    args = parse_args(sys.argv[1:])
 
-    options = {
-        'verbose': args.verbose,
-        'debug': args.debug,
-        'log': args.log,
-        'add': not args.no_smt_add,
-        'sub': not args.no_smt_sub,
-        'mul': not args.no_smt_mul,
-        'div': args.smt_div,
-        'divByConst': args.smt_div_by_const,
-        'modByConst': args.smt_mod_by_const,
-        'expByConst': args.smt_exp_by_const,
-        'timeout': args.solver_timeout_branching,
-    }
+    options = mk_options(args)
 
     if args.width is not None:
         options['max_width'] = args.width
@@ -456,6 +519,11 @@ def main() -> int:
             size = assign[1].strip()
             arrlen[name] = int(size)
 
+    # quick bytecode execution mode
+    if args.bytecode is not None:
+        run_bytecode(args.bytecode, args, options)
+        return 0
+
     #
     # compile
     #
@@ -486,20 +554,24 @@ def main() -> int:
 
             for contract in contracts:
                 hexcode = source_unit.bytecodes_runtime[contract]
-                srcmap = source_unit.srcmaps_runtime[contract]
-                srcs = {}
                 abi = source_unit.abis[contract]
                 methodIdentifiers = source_unit.hashes(contract)
 
                 funsigs = [funsig for funsig in methodIdentifiers if funsig.startswith(args.function)]
 
-                setup_sig = methodIdentifiers.get('setUp()')
-
                 if funsigs:
                     num_passed = 0
                     num_failed = 0
                     print(f'\nRunning {len(funsigs)} tests for {filename.short}:{contract}')
-                    setup_ex = setup(hexcode, abi, srcmap, srcs, args, setup_sig, options)
+
+                    setup_sigs = sorted([ (k,v) for k,v in methodIdentifiers.items() if k == 'setUp()' or k.startswith('setUpPlus(') ])
+                    (setup_name, setup_sig, setup_selector) = (None, None, None)
+                    if len(setup_sigs) > 0:
+                        (setup_sig, setup_selector) = setup_sigs[-1]
+                        setup_name = setup_sig.split('(')[0]
+                        if args.verbose >= 2 or args.debug: print(f'Running {setup_sig}')
+                    setup_ex = setup(hexcode, abi, setup_name, setup_sig, setup_selector, arrlen, args, options)
+
                     for funsig in funsigs:
                         funselector = methodIdentifiers[funsig]
                         funname = funsig.split('(')[0]
@@ -508,6 +580,7 @@ def main() -> int:
                             num_passed += 1
                         else:
                             num_failed += 1
+
                     print(f'Symbolic test result: {num_passed} passed; {num_failed} failed')
                     total_passed += num_passed
                     total_failed += num_failed
